/*****************************************************************************
 *                DigCC: a minimal C compiler for DigMIPS                    *
 ***************************************************************************** 
 *                                                                           *
 * Copyright (c) 2012 Christophe Alias                                       *
 *                                                                           *
 * This program is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU General Public License as published by the     *
 * Free Software Foundation; either version 2 of the License, or (at your    * 
 * option) any later version.                                                *
 *                                                                           *
 * This program is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General  *
 * Public License for more details.                                          *
 *                                                                           *
 * You should have received a copy of the GNU General Public License along   *
 * with this program; if not, write to the Free Software Foundation, Inc.,   *
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                 *
 *                                                                           *
 * Written by Christophe Alias, Christophe.Alias@ens-lyon.fr                 *
 *                                                                           *
 *****************************************************************************/
#include "CodeDigmips.h"

#include <iostream>  //For the pretty-printer
#include <sstream> //for to_string()
#include <stdlib.h>  //for exit()

#include "Register.h"
#include "Label.h"

#define RETURN_REGISTER 2
#define SPILL0 0
#define SPILL1 1
#define SP 6
#define ARP 7

using namespace std;

//
// Implementation of abstract methods
//

BitVector* 
CodeDigmips::get_read_register()
{
  return read_register;
}

BitVector* 
CodeDigmips::get_write_register()
{
  return write_register;
}

string 
CodeDigmips::get_label()
{
  return label;
}

int 
CodeDigmips::is_jump()
{
  return (instruction == DIGMIPS_BLE || 
	  instruction == DIGMIPS_J || 
	  instruction == DIGMIPS_JA);
}

int 
CodeDigmips::is_label()
{
  return instruction == DIGMIPS_LABEL;
}


CodeDigmips* 
CodeDigmips::clone()
{
  CodeDigmips* result = new CodeDigmips(); //empty
  result->instruction = instruction;

  result->r_dest = r_dest;
  result->r1 = r1;
  result->r2 = r2;

  result->shift = shift; 

  result->label = label; 
  result->asm_code = asm_code; 

  result->read_register = read_register->clone();
  result->write_register = write_register->clone();

  return result;
}


//pretty-printer
string 
CodeDigmips::to_string()
{
  string result;
  stringstream sout;

  /*
    sout << "Read = " << *read_register 
    << "\tWrite = " << *write_register
    << "\t";
  */

  if(instruction == DIGMIPS_LABEL)
    sout << label << ":";
  else
    sout << "  "; //Indentation

  if(instruction == DIGMIPS_ADD)
    sout << "add r" << r_dest << ", r" << r1 << ", r" << r2; 
  
  if(instruction == DIGMIPS_SUB)
    sout << "sub r" << r_dest << ", r" << r1 << ", r" << r2; 

  if(instruction == DIGMIPS_LD)
    {
      if(shift >= 0)
	sout << "ld r" << r_dest << ", [r" << r1 << " + " << shift << "]";
      else
	sout << "ld r" << r_dest << ", [r" << r1 << " - " << -1*shift << "]";
    }

  if(instruction == DIGMIPS_ST)
    {
      if(shift >= 0)
	sout << "st r" << r1 << ", [r" << r_dest << " + " << shift << "]";
      else
	sout << "st r" << r1 << ", [r" << r_dest << " - " << -1*shift << "]";
    }

  if(instruction == DIGMIPS_BLE)
    sout << "ble r" << r1 << ", r" << r2 << ", " << label;

  if(instruction == DIGMIPS_LDI)
    sout << "ldi r" << r_dest << ", " << constant;

  if(instruction == DIGMIPS_LDI_LO)
    sout << "ldi r" << r_dest << ", lo(" << label << ")";

  if(instruction == DIGMIPS_LDI_HI)
    sout << "ldi r" << r_dest << ", hi(" << label << ")";

  if(instruction == DIGMIPS_JA)
    sout << "ja r" << r1 << ", r" << r2;

  if(instruction == DIGMIPS_J)
    sout << "j " << label;

  if(instruction == DIGMIPS_PUSH_ALL)
    sout << "push_all";

  if(instruction == DIGMIPS_POP_ALL)
    sout << "pop_all";

  if(instruction == DIGMIPS_INLINE_ASM)
    sout << asm_code << "\t//inline asm";
  
  result = sout.str(); //convert back to string
  return result;
}


string 
CodeDigmips::to_string_dot()
{
  return to_string();
}


int
CodeDigmips::read_r1()
{
  return(((instruction == DIGMIPS_ADD) || 
	  (instruction == DIGMIPS_SUB) || 
	  (instruction == DIGMIPS_LD) ||
	  (instruction == DIGMIPS_ST) ||
	  (instruction == DIGMIPS_BLE) ||
	  (instruction == DIGMIPS_JA))
	 && r1 != SPILL0 && r1 != SPILL1
	 && r1 != ARP && r1 != SP);
}

int
CodeDigmips::read_r2()
{
  return(((instruction == DIGMIPS_ADD) || 
	  (instruction == DIGMIPS_ADD) ||
	  (instruction == DIGMIPS_SUB) ||
	  (instruction == DIGMIPS_BLE) ||
	  (instruction == DIGMIPS_JA))
	 && r2 != SPILL0 && r2 != SPILL1
	 && r2 != ARP && r2 != SP);
}

int
CodeDigmips::read_r_dest()
{
  return ((instruction == DIGMIPS_ST)
	  && r_dest != SPILL0 && r_dest != SPILL1
	  && r_dest != ARP && r_dest != SP);
}

int
CodeDigmips::write_r_dest()
{
  return(((instruction == DIGMIPS_ADD) || 
	  (instruction == DIGMIPS_SUB) ||
	  (instruction == DIGMIPS_LD) ||
	  (instruction == DIGMIPS_LDI) ||
	  (instruction == DIGMIPS_LDI_LO) || 
	  (instruction == DIGMIPS_LDI_HI))
	  && r_dest != SPILL0 && r_dest != SPILL1
	  && r_dest != ARP && r_dest != SP);
}


void digmips_translate(Cfg* cfg, PseudoCode* code)
{
  pseudo_code_t instruction = code->instruction;

  if(instruction == ADD)
    digmips_add(cfg,code->r_dest,code->r1,code->r2);

  if(instruction == SUB)
    digmips_sub(cfg,code->r_dest,code->r1,code->r2);

  if(instruction == MUL)
    {
      digmips_push_all(cfg);
      push(cfg,code->r1);
      push(cfg,code->r2);
      call(cfg,"__mul");
      //Call to function __mul() in runtime library
      copy(cfg,code->r_dest,RETURN_REGISTER);
      digmips_pop_all(cfg);
    }

  if(instruction == DIV)
    {
      digmips_push_all(cfg);
      push(cfg,code->r1);
      push(cfg,code->r2);
      call(cfg,"__div");
      //Call to function __div() in runtime library
      copy(cfg,code->r_dest,RETURN_REGISTER);
      digmips_pop_all(cfg);
    }

  if(instruction == MOV_REG_INT)
    digmips_ldi(cfg,code->r_dest,code->constant);

  if(instruction == COPY)
    copy(cfg,code->r_dest,code->r1); 

  if(instruction == LOAD)
    digmips_ld(cfg,code->r_dest,code->r1,code->shift);

  if(instruction == STORE)
    digmips_st(cfg,code->r_dest,code->shift,code->r1);

  if(instruction == JUMP)
    digmips_j(cfg,code->label_id);

  if(instruction == CJUMP)
    cjump(cfg,code->relation,code->r1,code->r2,code->label_id);

  if(instruction == CALL)
    call(cfg,code->label_id);

  if(instruction == RET)
    ret(cfg,code->constant);

  if(instruction == LABEL)
    digmips_label(cfg,code->label_id);

  if(instruction == PUSH_ALL)
    digmips_push_all(cfg);

  if(instruction == POP_ALL)
    digmips_pop_all(cfg);

  if(instruction == PUSH)
    push(cfg,code->r1);

  if(instruction == POP)
    pop(cfg,code->r_dest);

  if(instruction == LOAD_ARP)
    digmips_ld(cfg,code->r_dest,ARP,code->shift);

  if(instruction == STORE_ARP)
    digmips_st(cfg,ARP,code->shift,code->r1);

  if(instruction == ADD_SP_REG)
    digmips_add(cfg,SP,SP,code->r1);

  if(instruction == SUB_SP_REG)
    digmips_sub(cfg,SP,SP,code->r1);

  if(instruction == MOV_REG_ARP)
    copy(cfg,code->r_dest,ARP);

  if(instruction == MOV_REG_SP)
    copy(cfg,code->r_dest,SP);

  if(instruction == MOV_SP_REG)
    copy(cfg,SP,code->r1);

  if(instruction == PUSH_ARP)
    push_(cfg,ARP); //no possible spill (r0 = 1)

  if(instruction == POP_ARP)
    pop_(cfg,ARP); //no possible spill (r0 = 1)

  if(instruction == MOV_ARP_SP)
    copy_(cfg,ARP,SP); //no possible spill (r0 = 0)

  if(instruction == MOV_SP_ARP)
    copy_(cfg,SP,ARP); //no possible spill (r0 = 0)

  if(instruction == INLINE_ASM)
    digmips_inline_asm(cfg,code->asm_code);
}


void digmips_apply(Cfg* cfg, CodeDigmips* code, Allocation* a)
{
  //cout << to_string() << endl
  //     << "==>" << endl;

  //
  // 0) Special: push_all, pop_all: just produce the code.
  //

  if(code->instruction == DIGMIPS_PUSH_ALL)
    {
      digmips_final_push_all(cfg);
      return;
    }

  if(code->instruction == DIGMIPS_POP_ALL)
    {
      digmips_final_pop_all(cfg);
      return;
    }

  //
  // 1) Map temporaries to registers whenever it applies
  //

  int _r_dest = code->r_dest; //will be used later...
  int _r1 = code->r1; 
  int _r2 = code->r2; 

  if(code->read_r1() && a->assignment[_r1]->is_register)
    code->r1 = a->assignment[_r1]->reg;

  if(code->read_r2() && a->assignment[_r2]->is_register)
    code->r2 = a->assignment[_r2]->reg;

  if((code->read_r_dest() || code->write_r_dest()) 
     && a->assignment[_r_dest]->is_register)
    code->r_dest = a->assignment[_r_dest]->reg;

  //
  // 2) Insert spill code if required
  //

  if(code->read_r1() && a->assignment[_r1]->is_stack)
    {
      digmips_ld(cfg,0,ARP,a->assignment[_r1]->shift);
      code->r1 = 0; //set to r0
    }
  //argument 1 spilled? => ld r0,[ARP+shift]

  if(code->read_r2() && a->assignment[_r2]->is_stack)
    {
      digmips_ld(cfg,1,ARP,a->assignment[_r2]->shift);
      code->r2 = 1; //set to r1
    }
  //argument 2 spilled? => ld r1,[ARP+shift]

  if(code->read_r_dest() && a->assignment[_r_dest]->is_stack)
    {
      digmips_ld(cfg,1,ARP,a->assignment[_r_dest]->shift);
      code->r_dest = 1; //set to r1
    }
  //st and r_dest (read) spilled ? => ld r1,[ARP+shift]

  int code_write_r_dest = false;
  if(code->write_r_dest() && a->assignment[_r_dest]->is_stack)
    {
      code_write_r_dest = true;
      code->r_dest = 1;
    }
  //r_dest spilled? => set to r1

  cfg->add_node(code);
  //insert the current instruction (with patched r1, r2, r_dest)

  if(code_write_r_dest && a->assignment[_r_dest]->is_stack)
    digmips_st(cfg,ARP,a->assignment[_r_dest]->shift,1);
  //r_dest spilled ? => st r1,[ARP+shift]

  //spill->print_code();
  //cout << "---------" << endl;
}


//Pretty constructors
void digmips_add(Cfg* cfg, int r_dest, int r1, int r2)
{
  CodeDigmips* result = new CodeDigmips();
  result->instruction = DIGMIPS_ADD;
  result->r_dest = r_dest;
  result->r1 = r1;
  result->r2 = r2;

  result->read_register->insert(r1);
  result->read_register->insert(r2);
  result->write_register->insert(r_dest);

  cfg->add_node(result);
}

void digmips_sub(Cfg* cfg, int r_dest, int r1, int r2)
{
  CodeDigmips* result = new CodeDigmips();
  result->instruction = DIGMIPS_SUB;
  result->r_dest = r_dest;
  result->r1 = r1;
  result->r2 = r2;

  result->read_register->insert(r1);
  result->read_register->insert(r2);
  result->write_register->insert(r_dest);

  cfg->add_node(result);
}

void digmips_ld(Cfg* cfg, int r_dest, int r_base, int shift)
{
  CodeDigmips* result = new CodeDigmips();
  result->instruction = DIGMIPS_LD;
  result->r_dest = r_dest;
  result->r1 = r_base;
  result->shift = shift;

  result->read_register->insert(r_base);
  result->write_register->insert(r_dest);

  cfg->add_node(result);
}

void digmips_st(Cfg* cfg, int r_dest, int shift, int r_source)
{
  CodeDigmips* result = new CodeDigmips();
  result->instruction = DIGMIPS_ST;
  result->r_dest = r_dest;
  result->r1 = r_source;
  result->shift = shift;

  result->read_register->insert(r_source);
  result->read_register->insert(r_dest);

  cfg->add_node(result);
}

void digmips_ble(Cfg* cfg, int r1, int r2, string label)
{
  CodeDigmips* result = new CodeDigmips();
  result->instruction = DIGMIPS_BLE;
  result->r1 = r1;
  result->r2 = r2;
  result->label = label;

  result->read_register->insert(r1);
  result->read_register->insert(r2);

  cfg->add_node(result);
}

void digmips_ldi(Cfg* cfg, int r_dest, int constante)
{
  CodeDigmips* result = new CodeDigmips();
  result->instruction = DIGMIPS_LDI;
  result->r_dest = r_dest;
  result->constant = constante;

  result->write_register->insert(r_dest);

  cfg->add_node(result);
}

void digmips_ldi_lo(Cfg* cfg, int r_dest, string label)
{
  CodeDigmips* result = new CodeDigmips();
  result->instruction = DIGMIPS_LDI_LO;
  result->r_dest = r_dest;
  result->label = label;

  result->write_register->insert(r_dest);

  cfg->add_node(result);
}

void digmips_ldi_hi(Cfg* cfg, int r_dest, string label)
{
  CodeDigmips* result = new CodeDigmips();
  result->instruction = DIGMIPS_LDI_HI;
  result->r_dest = r_dest;
  result->label = label;

  result->write_register->insert(r_dest);

  cfg->add_node(result);
}

void digmips_ja(Cfg* cfg, int r1, int r2)
{
  CodeDigmips* result = new CodeDigmips();
  result->instruction = DIGMIPS_JA;
  result->r1 = r1;
  result->r2 = r2;
  result->label = "__d_u_m_m_y__"; //dummy value

  result->read_register->insert(r1);
  result->read_register->insert(r2);

  cfg->add_node(result);
}

void digmips_j(Cfg* cfg, string label)
{
  CodeDigmips* result = new CodeDigmips();
  result->instruction = DIGMIPS_J;
  result->label = label;
  cfg->add_node(result);
}

void digmips_label(Cfg* cfg, string label)
{
  CodeDigmips* result = new CodeDigmips();
  result->instruction = DIGMIPS_LABEL;
  result->label = label;
  cfg->add_node(result);
}

void digmips_push_all(Cfg* cfg)
{
  CodeDigmips* result = new CodeDigmips();
  result->instruction = DIGMIPS_PUSH_ALL;
  cfg->add_node(result);
}

void digmips_pop_all(Cfg* cfg)
{
  CodeDigmips* result = new CodeDigmips();
  result->instruction = DIGMIPS_POP_ALL;
  cfg->add_node(result);
}


void digmips_final_push_all(Cfg* cfg)
{
  digmips_ldi(cfg,0,3);
  //r0 = 3
  digmips_st(cfg,SP,0,3);
  //[SP] = r3
  digmips_st(cfg,SP,-1,4);
  //[SP-1] = r4
  digmips_st(cfg,SP,-2,5);
  //[SP-2] = r5
  digmips_sub(cfg,SP,SP,0);
  //SP = SP - 3
}


void digmips_final_pop_all(Cfg* cfg)
{
  digmips_ldi(cfg,0,3);
  //r0 = 3
  digmips_ld(cfg,5,SP,1);
  //r5 = [SP+1]
  digmips_ld(cfg,4,SP,2);
  //r4 = [SP+2]
  digmips_ld(cfg,3,SP,3);
  //r3 = [SP+3]
  digmips_add(cfg,SP,SP,0);
  //SP = SP + 3
}


void digmips_inline_asm(Cfg* cfg,string asm_code)
{
  CodeDigmips* result = new CodeDigmips(); //no read, no write
  result->instruction = DIGMIPS_INLINE_ASM;
  result->asm_code = asm_code;
  cfg->add_node(result);
}


void cjump(Cfg* cfg, test_t test, int r1, int r2, string label)
{
  if(test == LE)
    digmips_ble(cfg,r1,r2,label);

  if(test == GE)
    digmips_ble(cfg,r2,r1,label);

  if(test == L)
    {
      int r2_minus_1 = new_register();
      int r_one = new_register();
      digmips_ldi(cfg,r_one,1);
      digmips_sub(cfg,r2_minus_1,r2,r_one);
      digmips_ble(cfg,r1,r2_minus_1,label);
    }

  if(test == G)
    cjump(cfg,L,r2,r1,label);

  if(test == E)
    {
      string label_le = new_label("lower");
      string label_neq = new_label("not_equal");
      digmips_ble(cfg,r1,r2,label_le);
      digmips_j(cfg,label_neq);
      digmips_label(cfg,label_le);
      digmips_ble(cfg,r2,r1,label);
      digmips_label(cfg,label_neq);      
    }

  if(test == NE)
    {
      string label_le = new_label("lower");
      string label_eq = new_label("equal");
      digmips_ble(cfg,r1,r2,label_le);
      digmips_j(cfg,label);
      digmips_label(cfg,label_le);
      digmips_ble(cfg,r2,r1,label_eq);
      digmips_j(cfg,label);
      digmips_label(cfg,label_eq);      
    }
}


void copy(Cfg* cfg, int r_dest, int r_source)
{
  int r_zero = new_register();
  digmips_ldi(cfg,r_zero,0);
  digmips_add(cfg,r_dest,r_source,r_zero);
}

void copy_(Cfg* cfg, int r_dest, int r_source)
{
  digmips_ldi(cfg,0,0); //r0 = 0
  digmips_add(cfg,r_dest,r_source,0);
}

void push(Cfg* cfg, int r)
{
  int r_one = new_register();
  digmips_ldi(cfg,r_one,1);

  digmips_st(cfg,SP,0,r);
  //[SP] = r

  digmips_sub(cfg,SP,SP,r_one);
  //SP = SP - 1
}

void pop(Cfg* cfg, int r)
{
  int r_one = new_register();
  digmips_ldi(cfg,r_one,1);

  digmips_add(cfg,SP,SP,r_one);
  //SP = SP + 1

  digmips_ld(cfg,r,SP,0);
  //r = [SP]
}

void push_(Cfg* cfg, int r)
{
  digmips_st(cfg,SP,0,r);
  //[SP] = r
  digmips_ldi(cfg,0,1);
  //r0 = 1
  digmips_sub(cfg,SP,SP,0);
  //SP = SP - 1
}


void pop_(Cfg* cfg, int r)
{
  digmips_ldi(cfg,0,1);
  //r0 = 1

  digmips_add(cfg,SP,SP,0);
  //SP = SP + 1

  digmips_ld(cfg,r,SP,0);
  //r = [SP]
}

void call(Cfg* cfg, string label)
{
  string label_ret = new_label("return_point");
  int r_lo = new_register();
  int r_hi = new_register();

  digmips_ldi_lo(cfg,r_lo,label_ret);
  digmips_ldi_hi(cfg,r_hi,label_ret);
  push(cfg,r_lo);
  //push lo(label_ret)
  push(cfg,r_hi);
  //push hi(label_ret)
  digmips_j(cfg,label);
  //j label
  ((CodeDigmips*)cfg->node[cfg->node.size()-1])->write_register
    ->insert(RETURN_REGISTER);
  //function call -> result_register (r2) is written
  digmips_label(cfg,label_ret);
  //label_ret:
}

void ret(Cfg* cfg, int n)
{
  digmips_ldi(cfg,SPILL0,n+2);
  //r0 = n+2
  digmips_add(cfg,SP,SP,SPILL0);
  //SP = SP + (n+2)
  digmips_ld(cfg,SPILL0,SP,-1*(n+1));
  //r0 = [SP-(n+1)] (=hi)
  digmips_ld(cfg,SPILL1,SP,-1*n);
  //r0 = [SP-(n+1)] (=lo)
  digmips_ja(cfg,SPILL0,SPILL1);
  //RET

  // int r_lo = new_register();
  // int r_hi = new_register();
  // pop(cfg,r_hi);  
  // pop(cfg,r_lo);

  // int r_n = new_register();
  // digmips_ldi(cfg,r_n,n);  
  // digmips_add(cfg,SP,SP,r_n);  
  // //SP = SP + n

  // digmips_ja(cfg,r_hi,r_lo);
  // //RET

  ((CodeDigmips*)cfg->node[cfg->node.size()-1])->read_register
    ->insert(RETURN_REGISTER);
  //ret -> result_register (r2) is read (will be live-out)
}
