/*****************************************************************************
 *                DigCC: a minimal C compiler for DigMIPS                    *
 ***************************************************************************** 
 *                                                                           *
 * Copyright (c) 2012 Christophe Alias                                       *
 *                                                                           *
 * This program is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU General Public License as published by the     *
 * Free Software Foundation; either version 2 of the License, or (at your    * 
 * option) any later version.                                                *
 *                                                                           *
 * This program is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General  *
 * Public License for more details.                                          *
 *                                                                           *
 * You should have received a copy of the GNU General Public License along   *
 * with this program; if not, write to the Free Software Foundation, Inc.,   *
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                 *
 *                                                                           *
 * Written by Christophe Alias, Christophe.Alias@ens-lyon.fr                 *
 *                                                                           *
 *****************************************************************************/
#include "Dag.h" 

#include <stdlib.h> //for exit() and system()
#include <fstream> //for show_dot() file I/O

Dag::Dag(BasicBlock* basic_block)
{
  map<int,PseudoCode*> const_node; //node defining 'constant'
  map<PseudoCode*,int> node_rank;
  map<PseudoCode*,vector<PseudoCode*> > father;

  for(int i=0; i<basic_block->block.size(); i++)
    //For each instruction...
    {
      PseudoCode* instruction = (PseudoCode*)basic_block->block[i];
      vector<int> read_register = instruction->read_register->to_vector();
      vector<int> write_register = instruction->write_register->to_vector();

      //
      // CASE 1: r_dest = r1 op r2
      //
      if(instruction->is_binary())
	{
	  int r_dest = instruction->r_dest; 
	  int r1 = instruction->r1; 
	  int r2 = instruction->r2; 

	  if(instruction->instruction == STORE)
	    r2 = r_dest;

	  //
	  // Rhs expression already defined and alive?
	  //
	  if((node_reg.find(r1) != node_reg.end()) &&
	     (node_reg.find(r2) != node_reg.end()) &&
	     (!(instruction->has_side_effect())))
	    //Both operands already defined?, not a side effect instruction?
	    {
	      int simplified = false;
	      for(int j=0; j<father[node_reg[r1]].size(); j++)
		for(int k=0; k<father[node_reg[r2]].size(); k++)
		  if((father[node_reg[r1]][j] == father[node_reg[r2]][k]) &&
		     (father[node_reg[r1]][j]->instruction == 
		      instruction->instruction))
		    //For each common father of both operands...
		    {
		      node_reg[r_dest] = father[node_reg[r1]][j];
		      //=> reuse the father node 
		      //==> *common subexpression elimination here*
		      simplified = true;
		    }
	      if(simplified) continue;
	    }

	  //
	  // Expression does not exists => add a node...
	  //

	  if(node_reg.find(r1) == node_reg.end())
	    //Operand 1 not defined?
	    {
	      PseudoCode* leaf = pseudo_copy(r1,r1);
	      node.push_back(leaf); //define it
	      node_rank[leaf] = node.size()-1;
	      node_reg[r1] = leaf;
	      vector<int> empty_succ;
	      succ.push_back(empty_succ); //no successors yet
	    }

	  if(node_reg.find(r2) == node_reg.end())
	    //Operand 2 not defined?
	    {
	      PseudoCode* leaf = pseudo_copy(r2,r2);
	      node.push_back(leaf); //define it
	      node_rank[leaf] = node.size()-1;
	      node_reg[r2] = leaf;
	      vector<int> empty_succ;
	      succ.push_back(empty_succ); //no successors yet
	    }
  	  
	  //Add the node 'instruction'
	  node.push_back(instruction);		     
	  node_rank[instruction] = node.size()-1;
	  vector<int> instruction_succ;
	  instruction_succ.push_back(node_rank[node_reg[r1]]);
	  instruction_succ.push_back(node_rank[node_reg[r2]]);
	  succ.push_back(instruction_succ);
	  //Set successors to r1 and r2
	  
	  father[node_reg[r1]].push_back(instruction);
	  father[node_reg[r2]].push_back(instruction);
	  //Set fathers of r1 and r2
	  
	  if(instruction->is_write()) //avoid degenerate cases
	    node_reg[r_dest] = instruction;
	  //replace the definition site of r_dest with instruction
	}

      //
      // CASE 2: r_dest = op r1
      //
      if(instruction->is_unary())
	{
	  int r_dest = instruction->r_dest; 
	  int r1 = instruction->r1; 

	  //
	  // Rhs expression already defined and alive?
	  //
	  if((node_reg.find(r1) != node_reg.end()) &&
	     (!(instruction->has_side_effect())))
	    //operand already defined?
	    {
	      int simplified = false;
	      for(int j=0; j<father[node_reg[r1]].size(); j++)
		//For each father of the operand...
		if(father[node_reg[r1]][j]->instruction ==
		   instruction->instruction)
		     //Same operation?
		    {
		      node_reg[r_dest] = father[node_reg[r1]][j];
		      //=> reuse the father node 
		      //==> *common subexpression elimination here*
		      simplified = true;
		    }
	      if(simplified) continue;
	    }
	    
	  //
	  // Expression does not exists => add a node...
	  //

	  if(node_reg.find(r1) == node_reg.end())
	    //Operand 1 not defined?
	    { 
	      PseudoCode* leaf = pseudo_copy(r1,r1);
	      node.push_back(leaf); //define it
	      node_rank[leaf] = node.size()-1;
	      node_reg[r1] = leaf;
	      vector<int> empty_succ;
	      succ.push_back(empty_succ); //no successors yet
	    }

	  //Add the node 'instruction'
	  node.push_back(instruction);		     
	  node_rank[instruction] = node.size()-1;
	  vector<int> instruction_succ;
	  instruction_succ.push_back(node_rank[node_reg[r1]]);
	  succ.push_back(instruction_succ);
	  //Set successor to r1
	  
	  father[node_reg[r1]].push_back(instruction);
	  //Set fathers of r1 to current node
	  if(instruction->is_write())
	    node_reg[r_dest] = instruction;
	  //replace the definition site of r_dest with instruction
	}

      //
      // CASE 3: r_dest = r1
      //
      if(instruction->is_copy())
	{
	  int r1 = instruction->r1;
	  int r_dest = instruction->r_dest;

	  //
	  // Operand r1 already defined?
	  //
	  if(node_reg.find(r1) != node_reg.end())
	    node_reg[r_dest] = node_reg[r1];
	    //then, it defines also r_dest

	  //
	  // Operand r1 not defined?
	  //
	  if(node_reg.find(r1) == node_reg.end())
	    {
	      //Add the node r1
	      node.push_back(instruction); //define it
	      node_rank[instruction] = node.size()-1;
	      node_reg[r1] = instruction;
	      node_reg[r_dest] = instruction;
	      //it defines both r1 and r_dest
	      vector<int> empty_succ;
	      succ.push_back(empty_succ); //no successors yet
	    }    
	}

      //
      // CASE 4: r_dest = constante
      //
      if(instruction->is_ldi())
	{
	  int r_dest = instruction->r_dest;
	  int constant = instruction->constant;

	  if(const_node.find(constant) != const_node.end())
	    //Constant already defined?
	    node_reg[r_dest] = const_node[constant];
	  else
	    {
	      //Add the node 'constant'
	      node.push_back(instruction); //define it
	      node_rank[instruction] = node.size()-1;
	      node_reg[r_dest] = instruction;
	      const_node[constant] = instruction;
	      //it defines r_dest
	      vector<int> empty_succ;
	      succ.push_back(empty_succ); //no successors yet
	    }
	}    

      //
      // CASE 5: 0 read instructions, mostly side effect instructions
      //
      if(instruction->is_leaf())
	{
	  int r_dest = instruction->r_dest;

	  //Add the node 'leaf'
	  node.push_back(instruction); //define it
	  node_rank[instruction] = node.size()-1;
	  if(instruction->is_write())
	    node_reg[r_dest] = instruction;
	  //it defines r_dest
	  vector<int> empty_succ;
	  succ.push_back(empty_succ); //no successors
	}
    }

  //Finally, set node_def.
  for(map<int,PseudoCode*>::const_iterator it = node_reg.begin(); 
      it != node_reg.end(); ++it)
    //For each register...
    node_def[it->second].push_back(it->first);

  //
  // Relate side effect instructions in the (reverse) text order
  // (to force a schedule in the original order)
  //

  int previous_side_effect_instruction = -1;
  for(int i=0; i<node.size(); i++)
    if(node[i]->has_side_effect())
      //For each side-effect instruction...
      {
	if(previous_side_effect_instruction != -1)
	  succ[i].push_back(previous_side_effect_instruction);
	previous_side_effect_instruction = i;
      }

  //
  // Get the roots (nodes w/o predecessors)
  //

  for(int i=0; i<node.size(); i++)
    //Roots <-- all the nodes...
    roots.insert(i);

  for(int i=0; i<succ.size(); i++)
    for(int j=0; j<succ[i].size(); j++)
      //... then, remove each possible successor
      roots.erase(succ[i][j]);

  // cout << "Roots:" << endl;
  // for(set<int>::const_iterator it = roots.begin(); it != roots.end(); ++it)
  //   cout << *it << " ";
  // cout << endl;
  //DEBUG
}


vector<int> is_marked;
vector<vector<int> > is_marked_edge; //edges rank to be removed from succ.

void 
Dag::simplify_rec(int node_rank)
{
  if(is_marked[node_rank])
    return;

  is_marked[node_rank] = true;


  //
  // Simplify the operands first
  //
  for(int i=0; i<succ[node_rank].size(); i++)
    simplify_rec(succ[node_rank][i]);

  //
  // Then, simplify the node itself
  //
  
  // Simple algebraic simplifications

  int left_son,right_son;
  if(succ[node_rank].size() >= 2)
    {
      left_son = succ[node_rank][0];
      right_son = succ[node_rank][1];

      int is_add = (node[node_rank]->instruction == ADD);
      int is_mul = (node[node_rank]->instruction == MUL);
      int left_ldi = (node[left_son]->instruction == MOV_REG_INT);
      int left_const = node[left_son]->constant;
      int right_ldi = (node[right_son]->instruction == MOV_REG_INT);
      int right_const = node[right_son]->constant;

      // multiplication-related simplifications

      if(is_mul && left_ldi && right_ldi)
	// x = 2*3 ==> x = 6
	{
	  node[node_rank]->instruction = MOV_REG_INT;
	  node[node_rank]->constant = left_const*right_const; //do it

	  is_marked_edge[node_rank].push_back(0);
	  is_marked_edge[node_rank].push_back(1); 
	  //schedule both edges to be removed
	}

      if(is_mul && left_ldi && (left_const == 1) && !right_ldi)
	//x = 1*y ==> x = y
	{
	  node[node_rank]->instruction = COPY;
	  node[node_rank]->r1 = node[node_rank]->r2;

	  is_marked_edge[node_rank].push_back(0); 
	  //remove left edge
	}

      if(is_mul && !left_ldi && right_ldi && (right_const == 1))
	//x = y*1 ==> x = y
	{
	  node[node_rank]->instruction = COPY;
	  //node[node_rank]->r1 is already set in that case.

	  is_marked_edge[node_rank].push_back(1); 
	  //remove right edge
	}

      // addition-related simplifications

      if(is_add && left_ldi && right_ldi)
	// x = 2+3 ==> x = 5
	{
	  node[node_rank]->instruction = MOV_REG_INT;
	  node[node_rank]->constant = left_const+right_const; //do it

	  is_marked_edge[node_rank].push_back(0);
	  is_marked_edge[node_rank].push_back(1); 
	  //schedule both edges to be removed
	}

      if(is_add && left_ldi && (left_const == 0) && !right_ldi)
	//x = 0+y ==> x = y
	{
	  node[node_rank]->instruction = COPY;
	  node[node_rank]->r1 = node[node_rank]->r2;

	  is_marked_edge[node_rank].push_back(0); 
	  //remove left edge
	}

      if(is_add && !left_ldi && right_ldi && (right_const == 0))
	//x = y+0 ==> x = y
	{
	  node[node_rank]->instruction = COPY;
	  //node[node_rank]->r1 is already set in that case.

	  is_marked_edge[node_rank].push_back(1); 
	  //remove right edge
	}
    }
}

void 
Dag::simplify(BitVector* live_out)
{
  is_marked.resize(node.size());
  for(int i=0; i<node.size(); i++)
    is_marked[i] = false;
  //Initially, no nodes are marked.

  is_marked_edge.resize(node.size());
  //Initially, no edges are marked.

  //
  // 1) Algebraic simplifications
  //

  for(set<int>::const_iterator it = roots.begin(); it != roots.end(); ++it)
    //For each root node, simplify...
    simplify_rec(*it);

  //
  // 2) Remove edges
  //
  
  vector<vector<int> > new_succ;

  for(int i=0; i<node.size(); i++)
    for(int j=0; j<is_marked_edge[i].size(); j++)
      ;      


  //
  // 3) Remove dead-nodes
  //


}


void
Dag::print_dot(ostream& sout)
{
  sout << "digraph g {"<< endl;

  //Nodes
  for(int i=0; i<node.size(); i++)
    {
      sout << "  node" << i 
	   << "[label=\"" << node[i]->to_string_dot() << "\\n{";
      for(int j=0; j<node_def[node[i]].size(); j++)
	sout << "r" << node_def[node[i]][j] << " ";
      sout << "}\",shape=box];" << endl;
    }

  //Edges
  for(int i=0; i<succ.size(); i++)
    for(int j=0; j<succ[i].size(); j++)
      sout << "  node" << i << " -> node" << succ[i][j] << ";";
  sout << "}" << endl;
}


void 
Dag::show_dot()
{
  ofstream output_file;
  output_file.open("dag.dot");
  print_dot(output_file);
  output_file.close();
  //Produce the .dot

  system("dot -Tps dag.dot > dag.ps");
  system("evince dag.ps &");
  //Compile and display
}
